from string import ascii_uppercase

from django.db import models
from django.core.exceptions import ValidationError

from django.db import models


class ReusableAutoField(models.PositiveIntegerField):
    """A django auto field that can reuse deleted primary keys"""

    def get_next_available_id(self, model_cls, using=None):
        """
        Returns the next available id for the given model class.
        """
        all_ids = set(range(1, model_cls._default_manager.count()+1))
        used_ids = set(model_cls._default_manager.all().values_list('pk', flat=True))
        available_ids = all_ids - used_ids

        if available_ids:
            return min(available_ids)

        if used_ids:
            return max(used_ids) + 1

        return 1

    def get_default(self):
        """Returns the default value for this field"""

        return self.get_next_available_id(self.model)


# class Peg(models.Model):
#     """Represents a person's peg"""

#     peg_number = ReusableAutoField(primary_key=True, default=ReusableAutoField.get_default, editable=False)

#     def __str__(self):
#         return f"Peg {self.peg_number}"


class SectionValidator:
    """Validation class for the `section` field on the `member` model."""

    def __init__(self, max_value="ZZZ"):
        self.max_value = max_value.upper()
        self.alphabet_size = ord("Z") - ord("A") + 1

    def is_valid(self, section: str, team_sections: list[str]=None) -> bool:
        """Returns boolean if the section passed is valid."""
        section = section.upper()

        if not self._is_alphanumeric(section):
            return False

        if not self._is_in_alphabet(section[0]):
            return False

        if not self._is_in_range(section):
            return False

        if team_sections:
            if not self._is_unique(section, team_sections):
                return False

            if not self._is_not_adjacent(section, team_sections):
                return False

        return True

    def _is_alphanumeric(self, section: str) -> bool:
        """Returns boolean if all characters in the passed string are alphanumerical."""
        return all(c.isalnum() for c in section)

    def _is_in_alphabet(self, c) -> bool:
        """Returns boolean if the passed character is alphabetical."""
        return "A" <= c <= "Z"

    def _is_in_range(self, section) -> bool:
        """Returns boolean if the passed section less or equal to the max value."""
        section_value = self._section_value(section)
        max_value = self._section_value(self.max_value)

        return section_value <= max_value

    def _is_unique(self, section, team_sections) -> bool:
        """Returns boolean if the passed section is unique amongst `team_sections`."""
        return section not in team_sections

    def _is_not_adjacent(self, section, team_sections) -> bool:
        """Returns boolean if the passed section is not adjacent to any `team_sections`."""
        for team_section in team_sections:
            team_section_value = self._section_value(team_section)
            section_value = self._section_value(section)
            if abs(team_section_value - section_value) <= 1:
                return False

        return True

    def _section_value(self, section):
        """Returns the value of the passed section."""
        n = len(section)
        value = sum((ord(c) - ord("A") + 1) * self.alphabet_size ** (n - i - 1) for i, c in enumerate(section))
        return value


class Section(models.Model):
    """Represents a fishing area. Members can be assigned to a section,
    but no 2 teammates can be in the same or adjacent section."""

    name = models.CharField(max_length=3, unique=True, null=False)

    def clean(self):
        super().clean()

        validator = SectionValidator(max_value="ZZZ")
        if not validator.is_valid(self.name):
            raise ValidationError("Invalid section value.")

        self.name = self.name.upper()

    def __str__(self) -> str:
        return self.name

    @property
    def members(self):
        return Member.objects.filter(section=self)


def validate_team_size(value):
    if Member.objects.filter(team=value).count() >= 3:
        raise ValidationError('Team already has maximal amount of members (3)')


class Member(models.Model):
    """Represents a member of a team"""

    first_name = models.CharField(max_length=255)
    last_name = models.CharField(max_length=255)
    team = models.ForeignKey("Team", on_delete=models.SET_NULL, null=True, blank=True, validators=(validate_team_size,), related_name='members')
    peg_number = models.PositiveIntegerField(null=True, editable=True, unique=True)
    section = models.ForeignKey(to=Section, on_delete=models.SET_NULL, null=True, swappable=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # If the peg_number field is not set, we assign it the smallest
        # available positive integer, excluding any used values.
        if not self.peg_number:
            used_peg_numbers = Member.objects.exclude(id=self.id).exclude(peg_number=None).values_list("peg_number", flat=True)
            peg_numbers = set(range(1, Member.objects.count() + 1)) - set(used_peg_numbers)
            if peg_numbers:
                self.peg_number = min (peg_numbers)

    # def clean(self):
    #     super().clean()

    #     validator = SectionValidator(max_value="ZZZ")
    #     if not validator.is_valid(self.section):
    #         raise ValidationError("Invalid section value.")

    def __str__(self):
        return f"{self.first_name} {self.last_name} (team {self.team.team_number}))"

    @property
    def fullname(self) -> str:
        return f"{self.first_name} {self.last_name}"


BLOCKED_SECTION_LETTERS = ["i"]  # lowercase only, sections cannot be any of these letters#


class Team(models.Model):
    """Represents a team"""

    name = ReusableAutoField(primary_key=True, default=ReusableAutoField.get_default, editable=False)
    # section_letter = models.CharField(max_length=1, unique=True, choices=[
    #     (char, char) for char in ascii_uppercase if char.lower() not in BLOCKED_SECTION_LETTERS
    # ])

    def __str__(self):
        return f"Team {self.name}"

    @property
    def members(self) -> list[Member]:
        return Member.objects.filter(team=self)


# def validate_section_character(value):
#     """Validates the section character"""

#     value = value.upper()
#     banned_characters = ["I"]

#     if value in banned_characters:
#         raise ValidationError(f"The character <{value}> is a prohibited character.")

